import { EMPTY, merge } from 'rxjs';
import { catchError, ignoreElements, switchMap, withLatestFrom, } from 'rxjs/operators';
import { onStrategy } from '../utils/onStrategy';
import { strategyHandling } from '../utils/strategy-handling';
import { getTNode, notifyAllParentsIfNeeded, notifyInjectingParentIfNeeded, templateHandling, } from './utils';
/**
 * @internal
 *
 * A factory function that returns a map of projections to turn a notification of a Observable (next, error, complete)
 *
 * @param customNextContext - projection function to provide custom properties as well as override existing
 */
export function notificationKindToViewContext(customNextContext) {
    // @TODO rethink overrides
    return {
        suspense: (value) => {
            return {
                $implicit: undefined,
                // if a custom value is provided take it, otherwise assign true
                $suspense: value !== undefined ? value : true,
                $error: false,
                $complete: false,
            };
        },
        next: (value) => {
            return Object.assign({ $implicit: value, $suspense: false, $error: false, $complete: false }, customNextContext(value));
        },
        error: (error) => {
            return {
                $complete: false,
                $error: error,
                $suspense: false,
            };
        },
        complete: () => {
            return {
                $error: false,
                $complete: true,
                $suspense: false,
            };
        },
    };
}
export function createTemplateManager(config) {
    const { renderSettings, notificationToTemplateName, templateSettings, } = config;
    const { defaultStrategyName, strategies, cdRef: injectingViewCdRef, patchZone, parent, eRef, } = renderSettings;
    const tNode = parent
        ? getTNode(injectingViewCdRef, eRef.nativeElement)
        : false;
    let activeTemplate;
    const strategyHandling$ = strategyHandling(defaultStrategyName, strategies);
    const templates = templateHandling(templateSettings.viewContainerRef, patchZone);
    const viewContainerRef = templateSettings.viewContainerRef;
    const triggerHandling = config.templateTrigger$ || EMPTY;
    const getContext = notificationKindToViewContext(templateSettings.customContext);
    return {
        addTemplateRef: templates.add,
        // addTrigger: triggerHandling.next,
        nextStrategy: strategyHandling$.next,
        render(values$) {
            return values$.pipe(
            /* tslint:disable */
            // mergeWith(triggerHandling.trigger$ || EMPTY),
            /* tslint:enable */
            withLatestFrom(strategyHandling$.strategy$), 
            // Cancel old renders
            switchMap(([notification, strategy]) => {
                const kind = notification.kind;
                const value = notification.value;
                const templateName = notificationToTemplateName[kind](value, templates);
                const template = templates.get(templateName);
                const isNewTemplate = activeTemplate !== templateName;
                const notifyParent = isNewTemplate && parent;
                return merge(onStrategy(value, strategy, (v, work, options) => {
                    if (isNewTemplate) {
                        if (viewContainerRef.length > 0) {
                            viewContainerRef.clear();
                        }
                        if (template) {
                            templates.createEmbeddedView(templateName);
                        }
                    }
                    if (template) {
                        const context = getContext[kind](value);
                        const view = viewContainerRef.get(0);
                        Object.keys(context).forEach((k) => {
                            view.context[k] = context[k];
                        });
                        work(view, options.scope, notification);
                    }
                    activeTemplate = templateName;
                }
                // we don't need to specify any scope here. The template manager is the only one
                // who will call `viewRef#detectChanges` on any of the templates it manages.
                // whenever a new value comes in, any pre-scheduled work of this taskManager will
                // be nooped before a new work will be scheduled. This happens because of the implementation
                // of `StrategyCredential#behavior`
                ).pipe(notifyAllParentsIfNeeded(tNode, injectingViewCdRef, strategy, () => notifyParent)), notifyInjectingParentIfNeeded(injectingViewCdRef, strategy, isNewTemplate).pipe(ignoreElements()));
            }), catchError((e) => {
                console.error(e);
                return EMPTY;
            }));
        },
    };
}
//# sourceMappingURL=template-manager.js.map